import inspect
from typing import Type, List, Dict, Callable, Union

import elasticsearch_dsl as e
from django.db import models
from rest_framework import serializers

from django_es_drf.document_registry import RegistrationContext


def generate_extra_serializer_fields(
    document: Type[e.Document],
    model: Type[models.Model],
    serializer: Type[serializers.ModelSerializer],
    serializer_mapping: Union[
        Dict[
            Type[e.Field],
            Union[Callable[[e.Field], RegistrationContext], Type[serializers.Field]],
        ],
        Dict[
            str,
            Union[Callable[[e.Field], RegistrationContext], Type[serializers.Field]],
        ],
    ],
    included: List[str],
    excluded: List[str],
    prefix="",
):
    props = document._doc_type.mapping.properties.properties
    ctx = RegistrationContext(
        model=model,
        serializer=serializer,
        mapping=serializer_mapping,
        prefix=prefix,
        included=included,
        excluded=excluded,
    )

    extra_fields = generate_mapping(prefix, ctx, serializer, included, excluded, props)

    if not extra_fields:
        return serializer
    return type(f"{document.__name__}Serializer", (serializer,), extra_fields)


def generate_mapping(prefix, ctx, existing_serializer, included, excluded, props):
    serializer_fields = existing_serializer().fields if existing_serializer else {}
    new_mapping = {}
    for fld_name in props:
        prefixed_name = prefix + fld_name
        if fld_name in serializer_fields:
            continue
        if included and prefixed_name not in included:
            continue
        if excluded and prefixed_name in excluded:
            continue
        fld = props[fld_name]
        new_mapping[fld_name] = generate_field_mapping(
            prefixed_name, fld_name, fld, ctx
        )
    return new_mapping


def generate_field_mapping(prefixed_name, fld_name, fld, ctx):
    if fld._multi:
        multi = True
    else:
        multi = False
    es_field_callable = get_serializer_field_from_es_field(prefixed_name, fld, ctx)
    if inspect.isclass(es_field_callable) and issubclass(es_field_callable, e.Field):
        ret = es_field_callable()
    elif callable(es_field_callable):
        ret = es_field_callable(fld_name, fld, ctx)
    else:
        raise ValueError(
            f"Bad value for field {prefixed_name}: expecting DSL field or callable, got {es_field_callable}"
        )
    if not multi:
        return ret
    return serializers.ListField(child=ret)


def get_serializer_field_from_es_field(prefixed_name, fld, ctx):
    if prefixed_name in ctx.mapping:
        return ctx.mapping[prefixed_name]

    for m in type(fld).mro():
        if m in ctx.mapping:
            return ctx.mapping[m]
    raise KeyError(
        f"Do not have Serializer mapping for ES field of type {type(fld)}. Either specify it in the `serializer_mapping` parameter"
        f"or exclude it with `excluded`"
    )



import elasticsearch_dsl as e
from rest_framework import serializers


def object_builder(fld_name, fld, ctx, **kwargs):
    new_mapping = _build_mapping(fld_name, fld, ctx)
    return e.Object(properties=new_mapping, **kwargs)


def nested_builder(fld_name, fld, ctx, **kwargs):
    new_mapping = _build_mapping(fld_name, fld, ctx)
    return e.Nested(properties=new_mapping, **kwargs)


def _build_mapping(fld_name, fld, ctx):
    from django_es_drf.document_registry import RegistrationContext
    from django_es_drf.document_generator import generate_mapping

    nested_prefix = ctx.prefix + fld_name + "."
    nested_context = RegistrationContext(**{**ctx._asdict(), "prefix": nested_prefix})
    # TODO: mozny merge, pokud uz nested/object field existuje na dokumentu
    new_mapping = generate_mapping(
        nested_prefix, nested_context, {}, ctx.included, ctx.excluded, fld
    )
    return new_mapping


def serializer_object_builder(fld_name, fld, ctx, **kwargs):
    from django_es_drf.serializer_generator import generate_mapping
    from django_es_drf.document_registry import RegistrationContext

    nested_prefix = ctx.prefix + fld_name + "."
    nested_context = RegistrationContext(**{**ctx._asdict(), "prefix": nested_prefix})
    new_mapping = generate_mapping(
        nested_prefix, nested_context, None, ctx.included, ctx.excluded, fld._mapping
    )
    return type(f"{fld_name}Serializer", (serializers.Serializer,), new_mapping)()



import copy

from elasticsearch_dsl import Search
from rest_framework.filters import BaseFilterBackend


class DynamicSourceBackend(BaseFilterBackend):
    """Dynamic source backend."""

    def filter_queryset(self, request, queryset, view):
        if not isinstance(queryset, Search):
            return queryset

        if hasattr(view, "is_dynamic_source_filtering_enabled"):
            if not view.is_dynamic_source_filtering_enabled(request):
                return queryset

        # only in listing urls
        lookup_url_kwarg = view.lookup_url_kwarg or view.lookup_field
        if lookup_url_kwarg in view.kwargs:
            return queryset

        include = request.GET.get("_include", None)
        exclude = request.GET.get("_exclude", None)

        if hasattr(view, "source") and view.source:
            if not isinstance(view.source, dict):
                declared_source = {"includes": [*view.source], "excludes": []}
            else:
                declared_source = {
                    "includes": [],
                    "excludes": [],
                    **copy.deepcopy(view.source),
                }
        else:
            declared_source = {"includes": [], "excludes": []}

        if exclude:
            declared_source["excludes"].extend(x.strip() for x in exclude.split(","))

        if include:
            declared_source["includes"].extend(x.strip() for x in include.split(","))
        if not declared_source["includes"]:
            declared_source["includes"] = ["*"]

        return queryset.source(declared_source)

    def to_regex(self, pathspec):
        pathspec = pathspec.replace(".", r"\.")
        pathspec = pathspec.replace("*", r".*")
        return f"^{pathspec}$"



from typing import Type

from rest_framework.exceptions import ValidationError
from rest_framework.fields import empty
from rest_framework.serializers import (
    BaseSerializer,
    Serializer,
    ModelSerializer,
    as_serializer_error,
)
from rest_framework.utils.serializer_helpers import ReturnDict


class StrictModelSerializer(ModelSerializer):
    def run_validation(self, data=empty):
        validated_data = super().run_validation(data)
        if data is not empty:
            data_keys = set(data.keys())
            validated_keys = set(validated_data.keys())

            for fld in self.fields.values():
                if fld.read_only and fld.field_name in data_keys:
                    data_keys.remove(fld.field_name)

            if data_keys != validated_keys:
                raise ValidationError(
                    detail=as_serializer_error(
                        ValidationError(f"Unexpected keys {data_keys - validated_keys}")
                    )
                )
        return validated_data


class ESDocumentSerializer(BaseSerializer):
    django_serializer: Type[Serializer]

    def __init__(self, *args, django_serializer=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.django_serializer = django_serializer

    def to_internal_value(self, data):
        serializer = self.django_serializer(data=data)
        serializer.is_valid(raise_exception=True)
        return serializer.validated_data

    def to_representation(self, instance):
        # just return the document
        return instance

    def create(self, validated_data):
        document_class = self.context["view"].document
        # TODO: nested are not supported
        doc = document_class(**validated_data)
        doc.save()
        return doc

    def update(self, instance, validated_data):
        if self.partial:
            for k, v in validated_data.items():
                setattr(instance, k, v)
        else:
            # clear the instance
            doctype = type(instance)
            if doctype.DOCUMENT_ID_FIELD not in validated_data:
                validated_data[doctype.DOCUMENT_ID_FIELD] = instance.meta.id
            instance = doctype(meta={"id": instance.meta.id}, **validated_data)
        instance.save()
        return instance

    def get_initial(self):
        return super().get_initial() or {}

    @property
    def fields(self):
        return self.django_serializer().fields

    @property
    def data(self):
        ret = super().data
        if isinstance(ret, dict):
            return ReturnDict(ret, serializer=self)
        else:
            return ret

    def __iter__(self):
        # TODO: implement this to get better support for DRF debug interface
        return iter([])



import inspect
from typing import Type, List, Dict, Callable, Union

from django.db import models
import elasticsearch_dsl as e
from rest_framework import serializers
from rest_framework.serializers import ListSerializer

from django_es_drf.document_registry import RegistrationContext


def generate_extra_document_fields(
    document: Type[e.Document],
    model: Type[models.Model],
    serializer: Type[serializers.ModelSerializer],
    mapping: Union[
        Dict[
            Type[serializers.Field],
            Union[Callable[[serializers.Field], RegistrationContext], Type[e.Field]],
        ],
        Dict[
            str,
            Union[Callable[[serializers.Field], RegistrationContext], Type[e.Field]],
        ],
    ],
    included: List[str],
    excluded: List[str],
    prefix="",
):
    existing_mapping = document._doc_type.mapping
    ctx = RegistrationContext(
        model=model,
        serializer=serializer,
        mapping=mapping,
        prefix=prefix,
        included=included,
        excluded=excluded,
    )

    new_mapping = generate_mapping(
        prefix, ctx, existing_mapping, included, excluded, serializer
    )

    if not new_mapping:
        return document
    return type(document.__name__, (document,), new_mapping)


def generate_mapping(prefix, ctx, existing_mapping, included, excluded, serializer):
    new_mapping = {}
    if inspect.isclass(serializer):
        serializer = serializer()

    for fld_name, fld in serializer.fields.items():
        prefixed_name = prefix + fld_name
        if fld_name in existing_mapping:
            continue
        if included and prefixed_name not in included:
            continue
        if excluded and prefixed_name in excluded:
            continue

        new_mapping[fld_name] = generate_field_mapping(
            prefixed_name, fld_name, fld, ctx
        )
    return new_mapping


def generate_field_mapping(prefixed_name, fld_name, fld, ctx):
    if isinstance(fld, ListSerializer):
        multi = True
        fld = fld.child
    else:
        multi = False
    es_field_callable = get_es_field_from_serializer_field(prefixed_name, fld, ctx)
    if inspect.isclass(es_field_callable) and issubclass(es_field_callable, e.Field):
        return es_field_callable(multi=multi)
    elif callable(es_field_callable):
        return es_field_callable(fld_name, fld, ctx, multi=multi)
    else:
        raise ValueError(
            f"Bad value for field {prefixed_name}: expecting DSL field or callable, got {es_field_callable}"
        )


def get_es_field_from_serializer_field(prefixed_name, fld, ctx):
    if prefixed_name in ctx.mapping:
        return ctx.mapping[prefixed_name]

    for m in type(fld).mro():
        if m in ctx.mapping:
            return ctx.mapping[m]
    raise KeyError(
        f"Do not have ES mapping for {prefixed_name}: {type(fld)}. Either specify it in the `mapping` parameter"
        f"or exclude it with `excluded`"
    )



import json
from json import JSONEncoder
from elasticsearch_dsl import Document, AttrList, AttrDict
from elasticsearch_dsl.response import AggResponse


def to_plain_json(doc):
    return json.loads(es_dump(doc.to_dict()))


class ESEncoderClass(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Document):
            return obj.to_dict(skip_empty=False)
        if isinstance(obj, AttrList):
            return obj._l_
        if isinstance(obj, AttrDict):
            return obj._d_
        if isinstance(obj, AggResponse):
            return obj.to_dict()
        return super().default(obj)


def es_dump(obj, **kwargs):
    return json.dumps(obj, cls=ESEncoderClass, **kwargs)
